/*
 * rh_gem.c: VIA Technology RHINE seris Fast Ethernet MAC driver
 *
 * Copyright (c) 2002-2005 Masayuki Murayama.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * 3. Neither the name of the author nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#pragma	ident	"@(#)rh_gem.c	1.27 05/04/12"

/*
 Change log
 0.9.0	11/23/2002
	first release of gem based rh driver

 0.9.1  11/30/2002
	Added debugging code for rhine1 emulation on rhine2 chips.
	0.9.0 did not work with rhine1 chips because gem did not check
	tx buffer alignment. Gem was fixed.

 0.9.2  12/1/2002
	rh_set_rx_filter(): multcast hash table was fixed
	Issue MII_RESET on attach

	02/16/2003
	struct gc cleared. gem_get_packet_default added.

	02/17/2003 rh_get_stats fixed for rhine I.
 0.9.3  02/17/2003 release
 	04/12/2002 ddi_dma_sync added
 	04/26/2002 rh_tx_desc_write was fixed for type mismactch.
 		   calling ddi_get/put8 in rhattach fixed for type mismatch
		   multicast cam filtering added (not tested)
		   shape up source code
 0.9.4  04/27/2003 release
 0.9.5  05/18/2003 no interrupt implemented
 0.9.5  05/18/2003 release
 0.9.8
        11/10/2003 rh_mii_init() removed
 1.0.0  11/10/2003 release
		   burst size tuning on tx underflow/rx overflow implemented.
 1.0.1  03/07/2004 release
		   TDWBI workaround implemented. Tx errors gone!
		   Licensing changed to the BSD license from LGPL.
 1.0.2  04/01/2004 release
		   a workaround for VT6103 implemented in rh_mii_write.
 1.0.5  04/29/2004 release
		   tx interrput optimizatoin for rhine1
 1.0.7  05/29/2004 release
		   bus error while eeprom autoloading fixed.
		   RCR accessed in wrong size (long).
 1.0.8  06/06/2004 release
 1.0.12 08/09/2004 rh_set_rx_filter fixed not to stop the nic core.
 1.0.18 04/12/2005 ddi_put/get are used for descriptor access
 */

/*
 TODO:
	tuning on tx underflow.
 */

/*
 * System Header files.
 */
#include <sys/types.h>
#include <sys/conf.h>
#include <sys/debug.h>
#include <sys/stropts.h>
#include <sys/stream.h>
#include <sys/strlog.h>
#include <sys/kmem.h>
#include <sys/stat.h>
#include <sys/kstat.h>
#include <sys/vtrace.h>
#include <sys/dlpi.h>
#include <sys/strsun.h>
#include <sys/ethernet.h>
#include <sys/modctl.h>
#include <sys/errno.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/ddi_impldefs.h>

#include <sys/pci.h>
#include "mii.h"
#include "gem.h"
#include "rhinereg.h"

static char	ident[] = "via rhine nic driver v" VERSION;
char	_depends_on[] = {"misc/gld"};

/* Debugging support */
#ifdef DEBUG_LEVEL
static int rh_debug = DEBUG_LEVEL;
#define	DPRINTF(n, args)	if (rh_debug>(n)) cmn_err args
#else
#define	DPRINTF(n, args)
#endif

/*
 * Useful macros and typedefs
 */
#define	FALSE	(0)
#define	TRUE	(!FALSE)
#define	ONESEC	drv_usectohz(1*1000000)

#define	RH_WRITE_IMR(dp, imr)	{	\
	OUTW((dp), IMR, (imr));	\
	if (((struct rh_dev *)dp->private)->revid > 0x40) {	\
		OUTB((dp), MIMR, (imr) >> 16);	\
	}	\
}

#define	RH_WRITE_ISR(dp, isr)	{	\
	OUTW((dp), ISR, (isr));	\
	if (((struct rh_dev *)dp->private)->revid > 0x40) {	\
		OUTB((dp), MISR, (isr) >> 16);	\
	}	\
}

#define	RH_READ_ISR(dp, isr)	{	\
	isr = INW((dp), ISR);	\
	if (((struct rh_dev *)dp->private)->revid > 0x40) {	\
		isr |= INB((dp), MISR) << 16;	\
	}	\
}

#define	RH_KICK_TX(dp, cr)	OUTB((dp), CR+1, ((CR_TDMD1 | (cr)) >> 8))

#ifdef MAP_MEM
# define	FLUSH(dp)	INW(dp, ISR)
#else
# define	FLUSH(dp)
#endif

#define	STRUCT_COPY(a, b)	bcopy(&(b), &(a), sizeof(a))

/*
 * Our configuration
 */
#define	OUR_INTR_BITS	\
	(ISR_ABTI | ISR_NORBF | ISR_PKRACE | ISR_OVFI |	\
	 ISR_CNT | ISR_BE | ISR_RU | ISR_TU |	\
	 ISR_RXE | ISR_TXE | 	\
	 ISR_PTX | ISR_PRX)

#ifndef TX_BUF_SIZE
# define TX_BUF_SIZE	64
#endif
#ifndef TX_RING_SIZE
# ifdef PTX_TEST
#  define TX_RING_SIZE	16
# else
#  define TX_RING_SIZE	(TX_BUF_SIZE * 4)
# endif
#endif

#ifndef RX_BUF_SIZE
# define RX_BUF_SIZE	64
#endif
#ifndef RX_RING_SIZE
# define RX_RING_SIZE	RX_BUF_SIZE
#endif

static int	rh_tx_copy_thresh = 256;
static int	rh_rx_copy_thresh = 256;

/*
 * rhine chip state
 */
struct rh_dev {
	uint8_t		mac_addr[ETHERADDRL];
	uint16_t	cr;	/* soft copy of command reg. */
	uint8_t		revid;	/* chip revision id */
	uint8_t		tft;	/* tx fifo threshold */
	uint8_t		rft;	/* rx fifo threshold */
	uint8_t		maxdma;
	volatile uint32_t	imr;
	volatile uint32_t	imr_hw;
	boolean_t	td_unreliable;
#ifdef RESET_TEST
	int		reset_test;
#endif
#ifdef OPT_TX_INTR
	ddi_softintr_t	soft_id;
	boolean_t	imr_change_req;
#endif
};

/*
 * Supported chips
 */
struct chip_info {
	uint16_t	venid;
	uint16_t	devid;
	char		*name;
};

static struct chip_info rh_chiptbl[] = {
	{VID_VIA, DID_VT86C100A, "VT86C100A Rhine"},
	{VID_VIA, DID_VT3043, "DL10030/VT3043 Rhine"},
	{VID_VIA, DID_VT6102, "VT6102 Rhine-II"},
	{VID_VIA, DID_VT6105, "VT6105 Rhine-III"},
#ifdef CONFIG_VT6105M
	{VID_VIA, DID_VT6105M, "VT6105M Rhine-III"},
#endif
};
#define CHIPTABLESIZE   (sizeof(rh_chiptbl)/sizeof(struct chip_info))

/*
 * Macros to identify chip generation.
 */
#define	IS_RHINE1_VT86C100A_E(r)	((r) <= 0x4)
#define	IS_RHINE1_VT86C100A(r)		((r) < 0x20)
#define	IS_RHINE1(r)			((r) < 0x40)
#define	IS_RHINE2(r)			(0x40 <= (r) && (r) < 0x80)
#define	IS_RHINE2_OR_LATER(r)		(0x40 <= (r))
#define	IS_RHINE3(r)			(0x80 <= (r) && (r) < 0x90)
#define	IS_RHINE3_OR_LATER(r)		(0x80 <= (r))
#define	IS_RHINE3LOM(r)			(0x8a <= (r) && (r) < 0x90)
#define	IS_RHINE3M(r)			(0x90 <= (r))
#define	IS_RHINE3M_OR_LATER(r)		(0x90 <= (r))

#define	IS_VT6103(phyid)	\
	(((phyid) & 0xfffffff0) == 0x01018f20 && ((phyid) & 0xf) > 4)
/* ======================================================== */
 
/* mii operations */
static void  rh_mii_sync(struct gem_dev *);
static uint16_t  rh_mii_read(struct gem_dev *, int);
static void rh_mii_write(struct gem_dev *, int, uint16_t);

/* nic operations */
static int rh_reset_chip(struct gem_dev *);
static void rh_init_chip(struct gem_dev *);
static void rh_start_chip(struct gem_dev *);
static void rh_stop_chip(struct gem_dev *);
static void rh_set_media(struct gem_dev *);
static void rh_set_rx_filter(struct gem_dev *);
static void rh_get_stats(struct gem_dev *);
static int rh_init_mac_addr(struct gem_dev *);

/* descriptor operations */
static int rh_tx_desc_write(struct gem_dev *dp, uint_t slot,
		    ddi_dma_cookie_t *dmacookie, int frags, uint32_t flags);
static int rh_rx_desc_write(struct gem_dev *dp, uint_t slot,
		    ddi_dma_cookie_t *dmacookie, int frags);
static uint_t rh_tx_desc_stat(struct gem_dev *dp, int slot, int ndesc);
static uint_t rh_rx_desc_stat(struct gem_dev *dp, int slot, int ndesc);

static void rh_tx_desc_init(struct gem_dev *dp, int slot);
static void rh_rx_desc_init(struct gem_dev *dp, int slot);
static void rh_tx_desc_clean(struct gem_dev *dp, int slot);
static void rh_rx_desc_clean(struct gem_dev *dp, int slot);

/* interrupt handler */
static u_int rh_interrupt(struct gem_dev *dp);

/* ======================================================== */

/* mapping attributes */
/* Data access requirements. */
static struct ddi_device_acc_attr rh_dev_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC
};

/* On sparc, the buffers should be native endianness */
static struct ddi_device_acc_attr rh_buf_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_NEVERSWAP_ACC,	/* native endianness */
	DDI_STRICTORDER_ACC
};

static ddi_dma_attr_t rh_dma_attr_buf = {
	DMA_ATTR_V0,		/* dma_attr_version */
	0,			/* dma_attr_addr_lo */
	0xffffffffull,		/* dma_attr_addr_hi */
	0xffffffffull,		/* dma_attr_count_max */
	0,/* patched later */	/* dma_attr_align */
	0xffffffff,		/* dma_attr_burstsizes */
	1,			/* dma_attr_minxfer */
	0xffffffffull,		/* dma_attr_maxxfer */
	0xffffffffull,		/* dma_attr_seg */
	0,/* patched later */	/* dma_attr_sgllen */
	1,			/* dma_attr_granular */
	0			/* dma_attr_flags */
};

static ddi_dma_attr_t rh_dma_attr_desc = {
	DMA_ATTR_V0,		/* dma_attr_version */
	0,			/* dma_attr_addr_lo */
	0xffffffffull,		/* dma_attr_addr_hi */
	0xffffffffull,		/* dma_attr_count_max */
	16,			/* dma_attr_align */
	0xffffffff,		/* dma_attr_burstsizes */
	1,			/* dma_attr_minxfer */
	0xffffffffull,		/* dma_attr_maxxfer */
	0xffffffffull,		/* dma_attr_seg */
	1,			/* dma_attr_sgllen */
	1,			/* dma_attr_granular */
	0			/* dma_attr_flags */
};

/* ======================================================== */
/*
 * HW manupilation routines
 */
/* ======================================================== */

static int
rh_reset_chip(struct gem_dev *dp)
{
	int		i;
	int		ret;
	struct rh_dev	*lp = (struct rh_dev *)dp->private;

	DPRINTF(2, (CE_CONT, "!%s: rh_reset_chip: called, time:%d, cr:%b",
		dp->name, ddi_get_lbolt(),
		lp->cr, CR_BITS));

	ret = 0;
#ifdef notdef /* DEBUG_LEVEL */
	/* white a dummy mac address to ensure eeprom autoloading works */
	for (i = 0; i < ETHERADDRL; i++) {
		OUTB(dp, PAR + i, i);
	}
#endif
	/* invalidate a software cache of current mac address */
	bzero(lp->mac_addr, sizeof(lp->mac_addr));

	/* clear interrupt mask */
	RH_WRITE_IMR(dp, 0);

	/* Reset the chip. */
	lp->cr = 0;
	OUTB(dp, CR + 1, (CR_SRST >> 8) | INB(dp, CR + 1));

	if (IS_RHINE1(lp->revid)) {
		/* extra wait time for Rhine-I chips */
		drv_usecwait(10000);
	}

	i = 1;
	drv_usecwait(10);
	while ((INB(dp, CR + 1) & (CR_SRST >> 8)) != 0) {
		drv_usecwait(10);
		if (i++ > 1000) {
			cmn_err(CE_WARN, "%s: rh_reset_chip: timeout",
				dp->name);

			if (IS_RHINE1_VT86C100A(lp->revid)) {
				ret = -1;
				break;
			}

			/* Use force software reset bit in MISC register */

			OUTW(dp, MISC, MISC_FORSRST);
			drv_usecwait(2000);
			break;
		}
	}

	drv_usecwait(5000);

	/* clear interrupt mask again */
	RH_WRITE_IMR(dp, 0);

	/*
	 * Initialize MII logic
	 */
	OUTB(dp, MIICR, 0);

	/* wait for midle */
	if (IS_RHINE1_VT86C100A_E(lp->revid)) {
		/*
		 * turn off MSRCEN
		 */
		OUTB(dp, MIIADR, 1);
		drv_usecwait(1000);

		/* turn on MAUTO */
		OUTB(dp, MIICR, MIICR_MAUTO);
		i = 0;
		while ((INB(dp, MIIADR) & MIIADR_MDONE) == 0) {
			if (i++>100000) {
				break;
			}
			drv_usecwait(10);
		}

		/*
		 * as soon as MDONE is on,
		 * this is the right time to turn off MAUTO
		 */
		OUTB(dp, MIICR, 0);
	}
	else {
		i = 0;
		while ((INB(dp, MIIADR) & MIIADR_MIDLE) == 0) {
			if (i++>1000) {
				break;
			}
			drv_usecwait(10);
		}
	}
	DPRINTF(2, (CE_CONT, "!rh_reset_chip: mauto stopped in %d uS", i*10));

	return ret;
}

static void
rh_init_chip(struct gem_dev *dp)
{
	int		i;
	u_int		val;
	struct rh_dev	*lp = (struct rh_dev *)dp->private;

	DPRINTF(2, (CE_CONT, "!%s: rh_init_chip: called at time:%d",
		dp->name, ddi_get_lbolt()));

	/* PAR: do nothing */

	/* RCR: disable all rx filter */
	OUTB(dp, RCR, lp->rft << RCR_RRFT_SHIFT);

	/* TCR */
	OUTB(dp, TCR, (lp->tft << TCR_RTFT_SHIFT) | TCR_LB_NORMAL | TCR_OFSET);

	/* CR: don't touch  */

	/* ISR: don't touch */

	/* IMR & MIMR */
	lp->imr = 0;
	lp->imr_hw = lp->imr;
	RH_WRITE_IMR(dp, lp->imr_hw);

	/* Curr Rx Desc Addr */
	/* This seems to reset rx desc pointer. */
	OUTL(dp, CRDA, (uint32_t)dp->rx_ring_dma);

	/* Curr Tx Desc Addr */
	/* This seems to reset tx desc pointer. */
	OUTL(dp, CTDA, (uint32_t)dp->tx_ring_dma);
	if (IS_RHINE3M_OR_LATER(lp->revid)) {
		/* it has 8 transmit queues, but they aren't used. */
		for (i = 1; i < 8; i++) {
			OUTL(dp, CTDA + i*4,
				((uint32_t)dp->tx_ring_dma) +
				TX_RING_SIZE*sizeof(struct tx_desc)*i);
		}
	}

	/* BCR */
	OUTW(dp, BCR,
		(lp->tft << BCR_CTFT_SHIFT) |
		(lp->rft << BCR_CRFT_SHIFT) |
		lp->maxdma);

	/* CFGB */
	val = INB(dp, CFGB);
	if (IS_RHINE1(lp->revid)) {
		/* Disable queue packet*/
		val |= CFGB_QPKTDIS;
	}
	else {
		/* Enable queue packet*/
		val &= ~CFGB_QPKTDIS;
		OUTB(dp, WOLCG_SET, WOLCG_SAM | WOLCG_SAB);
	}
	OUTB(dp, CFGB, val);

	/* CFGD */
	val = INB(dp, CFGD);
	val &= ~(CFGD_CRANDOM | CFGD_CAP | CFGD_MBA | CFGD_BAKOPT);
	if (IS_RHINE1(lp->revid)) {
		/* Disable MII status change interrupt as it doesn't work */
		val &= ~CFGD_GPIOEN;
	}
	OUTB(dp, CFGD, val);

	if (IS_RHINE3_OR_LATER(lp->revid)) {
		/* RhineIII flow control */
		OUTB(dp, FCR1,
			FCR1_TXLOWAT_24 | FCR1_TXHIWAT_48 | FCR1_XONOFF_EN);
		OUTW(dp, PAUSE_TIMER, 0xffff);
		OUTB(dp, FCR0, min(RX_RING_SIZE, 255));
	}

	DPRINTF(2, (CE_CONT, "!%s: txthr:%d rxthr:%d maxdma:%d",
		dp->name, lp->tft, lp->rft, lp->maxdma));
}

static void
rh_start_chip(struct gem_dev *dp)
{
	struct rh_dev	*lp = (struct rh_dev *)dp->private;

	DPRINTF(2, (CE_CONT, "!%s: rh_start_chip: called at time:%d",
		dp->name, ddi_get_lbolt()));

	rh_set_media(dp);

	/* enable interrupt */
	lp->imr =
#ifdef OPT_TX_INTR
		  IS_RHINE1(lp->revid) ? (OUR_INTR_BITS & ~ISR_PTX) :
#endif
		  OUR_INTR_BITS;
	if (lp->revid > 0x40 && lp->revid < 0x90) {
		lp->imr |= MISR_TDWBI << 16;
	}

	lp->imr_hw = lp->imr;
	if ((dp->misc_flag & GEM_NOINTR) == 0) {
		RH_WRITE_IMR(dp, lp->imr_hw);
	}

	/* Kick TX and RX */
	lp->cr |= CR_TXON | CR_RXON | CR_DPOLL | CR_STRT;
	OUTW(dp, CR, lp->cr);
}

static void
rh_stop_chip(struct gem_dev *dp)
{
	struct rh_dev	*lp = (struct rh_dev *)dp->private;

	DPRINTF(2, (CE_CONT, "!%s: rh_stop_chip: called at time:%d",
		dp->name, ddi_get_lbolt()));

	/* Disable interrupts by clearing the interrupt mask */
	lp->imr = 0;
	lp->imr_hw = lp->imr;
	if ((dp->misc_flag & GEM_NOINTR) == 0) {
		RH_WRITE_IMR(dp, lp->imr_hw);
	}

	/* Stop Tx and Rx processes in the chip. */
	lp->cr &= ~(CR_TXON | CR_RXON | CR_DPOLL | CR_STRT);
	OUTW(dp, CR, lp->cr | CR_STOP);

	/* Note: No need to call rh_reset_chip() to reset Tx/Rx desc pointers */
}

#ifdef DEBUG_LEVEL
static uint16_t
rh_read_eeprom(struct gem_dev *dp, uint_t offset)
{
	int		i;
	uint16_t	ret;
	uint8_t		chip_select;
	uint8_t		di;
	uint8_t		cfga_saved;

#define	RH_EEPROM_DELAY(dp)	{INB(dp, EECSR);INB(dp, EECSR);}
#define	EE93C46_READ	6

	/* enable eeprom direct programing */
	cfga_saved = INB(dp, CFGA);
	OUTB(dp, CFGA, cfga_saved | CFGA_EELOAD);
	RH_EEPROM_DELAY(dp);

	/* ensure de-assert chip select */
	chip_select = EECSR_DPM;
	OUTB(dp, EECSR, chip_select);
	RH_EEPROM_DELAY(dp);

	/* assert chip select */
	chip_select |= EECSR_ECS;
	OUTB(dp, EECSR, chip_select);
	RH_EEPROM_DELAY(dp);

	/* make a read command for eeprom */
	offset = (offset & 0x3f) | (EE93C46_READ << 6);

	for (i = 10; i >= 0; i--) {
		/* send 1 bit */
		di = ((offset >> i) & 1) << EECSR_EDI_SHIFT;

		OUTB(dp, EECSR, chip_select | di);
		RH_EEPROM_DELAY(dp);

		OUTB(dp, EECSR, chip_select | di | EECSR_ECK);
		RH_EEPROM_DELAY(dp);
	}

	OUTB(dp, EECSR, chip_select);
	RH_EEPROM_DELAY(dp);

	/* get the reply and construct a 16bit value */
	ret = 0;
	for (i = 0; i < 16; i++) {
		/* Get 1 bit */
		OUTB(dp, EECSR, chip_select | EECSR_ECK);
		RH_EEPROM_DELAY(dp);

		ret = (ret << 1)
		    | ((INB(dp, EECSR) >> EECSR_EDO_SHIFT) & 1);

		OUTB(dp, EECSR, chip_select);
		RH_EEPROM_DELAY(dp);
	}

	/* negate chip_select */
	OUTB(dp, EECSR, EECSR_DPM);
	RH_EEPROM_DELAY(dp);

	OUTB(dp, EECSR, 0);
	RH_EEPROM_DELAY(dp);

	/* disable eeprom direct programming */
	OUTB(dp, CFGA, cfga_saved);
	RH_EEPROM_DELAY(dp);

	return ret;
}

static void
rh_eeprom_dump(struct gem_dev *dp)
{
	int		i;
	uint16_t	prom[0x10];

	for (i = 0; i < 0x10; i++) {
		prom[i] = rh_read_eeprom(dp, i);
	}

	cmn_err(CE_CONT, "!%s: eeprom dump:", dp->name);
	for (i = 0; i < 0x10; i += 4) {
		cmn_err(CE_CONT, "!0x%02x: 0x%04x 0x%04x 0x%04x 0x%04x", 
			i, prom[i], prom[i + 1], prom[i + 2], prom[i + 3]);
	}
}
#endif /* DEBUG_LEVEL */

static int
rh_attach_chip(struct gem_dev *dp)
{
	int		i;
	uint_t		val;
	uint8_t		*mac;
	struct rh_dev	*lp = (struct rh_dev *)dp->private;

	DPRINTF(2, (CE_CONT, "!%s: rh_attach_chip: called at time:%d",
		dp->name, ddi_get_lbolt()));

	mac = dp->dev_addr.ether_addr_octet;

	if (gem_get_mac_addr_conf(dp)) {
		/* the mac address has been specified by rh.conf */
		for (i = 0; i < ETHERADDRL; i++) {
			OUTB(dp, PAR + i, mac[i]);
		}
	}
	else {
		/*
		 * XXX - Don't read mac address from eeprom, it doesn't
		 * work with built-in rhine II core in VT8235 south bridge.
		 * We can read factory mac address from PAR reg after a reset.
		 */
		for (i = 0; i < ETHERADDRL; i++) {
			mac[i]= INB(dp, PAR + i);
		}
	}

	bcopy(mac, lp->mac_addr, ETHERADDRL);

#ifdef DEBUG_LEVEL
	rh_eeprom_dump(dp);
#endif /* DEBUG_LEVEL */

	/* read default address of PHY in eeprom */
	dp->mii_phy_addr = INB(dp, MPHY) & PHYADR_PHYAD;

#if 0
	dp->rxthr = 64;		/* rh-1.0.0 default */
#endif
	if (dp->rxthr <= 32) {
		lp->rft = RRFT_32;
	} else if (dp->rxthr <= 64) {
		lp->rft = RRFT_64;
	} else if (dp->rxthr <= 128) {
		lp->rft = RRFT_128;
	} else if (dp->rxthr <= 256) {
		lp->rft = RRFT_256;
	} else if (dp->rxthr <= 512) {
		lp->rft = RRFT_512;
	} else if (dp->rxthr <= 768) {
		lp->rft = RRFT_768;
	} else if (dp->rxthr <= 1024) {
		lp->rft = RRFT_1024;
	} else {
		lp->rft = RRFT_SF;
	}

	/* setup initial tft */
	if (IS_RHINE1(lp->revid)) {
		dp->txthr = max(dp->txthr, ETHERMAX); /* rh-1.0.0 default */
	}
	else if (IS_RHINE2(lp->revid)) {
		dp->txthr = max(dp->txthr, 512);
	}

	if (dp->txthr <= 128) {
		lp->tft = RTFT_128;
	} else if (dp->txthr <= 256) {
		lp->tft = RTFT_256;
	} else if (dp->txthr <= 512) {
		lp->tft = RTFT_512;
	} else if (dp->txthr <= 1024) {
		lp->tft = RTFT_1024;
	} else {
		lp->tft = RTFT_SF;
	}

	dp->txmaxdma = max(128, dp->txmaxdma);

	if (dp->txmaxdma < 32) {
		lp->maxdma = BCR_DMA_32;
	} else if (dp->txmaxdma < 64) {
		lp->maxdma = BCR_DMA_64;
	} else if (dp->txmaxdma < 128) {
		lp->maxdma = BCR_DMA_128;
	} else if (dp->txmaxdma < 256) {
		lp->maxdma = BCR_DMA_256;
	} else if (dp->txmaxdma < 512) {
		lp->maxdma = BCR_DMA_512;
	} else if (dp->txmaxdma < 1024) {
		lp->maxdma = BCR_DMA_1024;
	} else {
		lp->maxdma = BCR_DMA_NOLIMIT;
	}

#if 1
{
	clock_t	tim;
	int	i;
	i = 100000;
#define	TEST_COUNT	300
	tim = ddi_get_lbolt();

	/* OUTB(dp, GMCR, INB(dp, GMCR) | GMCR_TMR1US); */

	OUTW(dp, SOFT_TMR0, TEST_COUNT);
	/* OUTB(dp, MIMR, INB(dp, MIMR) | MIMR_TIM0IM); */
	OUTW(dp, MISC, INW(dp, MISC) | MISC_Tm0EN);
	while ((INB(dp, MISR) & MISR_TIM0I) == 0) {
		if (--i == 0) {
			cmn_err(CE_CONT, "%s: soft_timer0 timeout", dp->name);
			break;
		}
		drv_usecwait(100);
	}
	tim = (ddi_get_lbolt() - tim);

	if (i > 0) {
		cmn_err(CE_CONT, "%s: soft_timer0 resolution in 10mS: %d",
			dp->name, tim > 0 ? TEST_COUNT/tim: 9999999);
	}
	/* OUTL(dp, CR_CLR, CR_TM0EN); */
	
}
#endif
	return 0;
}

static uint_t
rh_mcast_hash(struct gem_dev *dp, uint8_t *addr)
{
	/* hash key is higher 6 bits of msb in crc */
	return gem_ether_crc_be(addr) >> (32 - 6);
}

static void
rh_set_rx_filter(struct gem_dev *dp)	
{
	uint32_t	mode;
	int		i, j;
	uint32_t	mhash[2];
	uint8_t		*mac;
	struct rh_dev	*lp = (struct rh_dev *)dp->private;

	DPRINTF(2, (CE_CONT, "!%s: rh_set_rx_filter: called at time:%d, "
		"active:%d, rxmode:%b, cr:%b",
		dp->name, ddi_get_lbolt(),
		dp->nic_active, dp->rxmode, "\020\002AllMulti\001Promisc",
		lp->cr, CR_BITS));

	mode = RCR_AB;	/* broadcast */
	mhash[0] = mhash[1] = 0;
	if ((dp->rxmode & RXMODE_PROMISC) != 0) {
		mode |=RCR_PRO	/* promiscous */
                     | RCR_AM;	/* muticast */
		mhash[0] = mhash[1] = 0xffffffff;
	}
	else if ((dp->rxmode & RXMODE_ALLMULTI) != 0 ||
				dp->mc_count > MULTICAST_CAM_SIZE) {
		mode |= RCR_AM;	/* multicast */
		mhash[0] = mhash[1] = 0xffffffff;
	}
#ifdef CONFIG_VT6105M
	else if (IS_RHINE3M_OR_LATER(lp->revid) && 
		 dp->mc_count <= MULTICAST_CAM_SIZE && dp->mc_count > 0) {
		/* use multicast cam. I guess multicast hash should be zero */
		mode |= RCR_AM;	/* multicast */
	}
#endif
	else if (dp->mc_count > 0) {
		mode |= RCR_AM;	/* multicast */

		/* make a hash table */
		for (i = 0; i < dp->mc_count; i++) {
			j = dp->mc_list[i].hash;
			mhash[j / 32] |= 1 << (j % 32);
		}
	}

#ifdef notdef
	/* need to disable rx engine before changing rx filter */
	if ((lp->cr & CR_RXON) != 0) {
		OUTW(dp, CR, lp->cr & ~CR_RXON);
	}
#else
	/* temporally disable rx filter before chaining it */
	OUTB(dp, RCR, (lp->rft << RCR_RRFT_SHIFT) | RCR_SEP | RCR_AR);
#endif
	/* set mac address */
	mac = &dp->cur_addr.ether_addr_octet[0];
	if (bcmp(lp->mac_addr, mac, ETHERADDRL) != 0) {
		for (i = 0; i < ETHERADDRL; i++) {
			OUTB(dp, PAR + i, mac[i]);
			lp->mac_addr[i] = mac[i];
		}
	}
#ifdef CONFIG_VT6105M
	if (IS_RHINE3M_OR_LATER(lp->revid)) {
		uint32_t	cammask;
		/*
		 * VT6105M has 32 entry cams for multicast filtering
		 */
		cammask = 0;

		/* enable cam port for writing multicast address*/
		OUTB(dp, CAMCTRL, CAMCTRL_EN);

		if (dp->mc_count <= MULTICAST_CAM_SIZE) {
			for (i = 0; i < dp->mc_count; i++) {
				/* write cam index */
				OUTB(dp, CAMADDR, i);

				/* write a multicast address into cam */
				mac = dp->mc_list[i].addr.ether_addr_octet;
				for (j = 0; j < ETHERADDRL; j++) {
					OUTB(dp, MAR + j, mac[j]);
				}
				drv_usecwait(10);

				/* issue a cam write command */
				OUTB(dp, CAMCTRL, CAMCTRL_EN | CAMCTRL_WR);
				drv_usecwait(10);

				cammask |= 1 << i;
			}
		}

		/* setup multicast cam mask */
		OUTL(dp, CAMMASK, cammask);
		OUTB(dp, CAMCTRL, 0);

		/* initialize VLAN cam mask */
		OUTB(dp, CAMCTRL, CAMCTRL_EN | CAMCTRL_VLAN);
		OUTL(dp, CAMMASK, 0);
		OUTB(dp, CAMCTRL, 0);
	}
#endif /* CONFIG_VT6105M */

	if ((mode & RCR_AM) != 0) {
		/* need to set multicast hash table */
		for (i = 0; i < 2; i++) {
			OUTL(dp, MAR + i*4, mhash[i]);
		}
	}

	/* update new rx filter mode */
	OUTB(dp, RCR, (lp->rft << RCR_RRFT_SHIFT) | mode | RCR_SEP | RCR_AR);
#ifdef notdef
	/* kick potentially stopped rx engine 06/19/2002 */
	if ((lp->cr & CR_RXON) != 0) {
		OUTW(dp, CR, lp->cr);
	}
#endif
}

static void
rh_set_media(struct gem_dev *dp)
{
	u_int		val;
	struct rh_dev	*lp = (struct rh_dev *)dp->private;

	DPRINTF(2, (CE_CONT, "!%s: rh_set_media: called at time:%d, "
		"active:%d, cr:%b",
		dp->name, ddi_get_lbolt(),
		dp->nic_active, lp->cr, CR_BITS));

	if (!dp->nic_active) {
		return;
	}

	/*
	 * Notify current duplex mode to mac
	 */

	/* need to disable rx and tx engine before changing duplex mode */
	if ((lp->cr & (CR_RXON | CR_TXON)) != 0) {
		OUTW(dp, CR, lp->cr & ~(CR_RXON | CR_TXON));
	}

	if (dp->full_duplex) {
		lp->cr |= CR_FDX;
	}
	else {
		lp->cr &= ~CR_FDX;
	}
	OUTB(dp, CR+1, lp->cr >> 8);

	/* kick rx and tx again */
	if ((lp->cr & (CR_RXON | CR_TXON)) != 0) {
		OUTW(dp, CR, lp->cr);
	}

	/* Set up flow control register */
	if (IS_RHINE2(lp->revid)) {
		/* VT6102 have partial flow control */
		val = INW(dp, MISC) & ~(MISC_FDXRFEN | MISC_HDXFEN);
		switch (dp->flow_control) {
		case FLOW_CONTROL_SYMMETRIC:
		case FLOW_CONTROL_RX_PAUSE:
			val |= MISC_FDXRFEN;
			break;
#ifdef NEVER
		case FLOW_CONTROL_NONE:
			if (!dp->full_duplex) {
				val |= MISC_HDXFEN;
			}
			break;
#endif
		}
		OUTW(dp, MISC, val);
	}	
	else if (IS_RHINE3_OR_LATER(lp->revid)) {
		val = INB(dp, FCR1) &
			~(FCR1_FDFCTX_EN | FCR1_FDFCRX_EN | FCR1_HDFC_EN);

		switch (dp->flow_control) {
		case FLOW_CONTROL_RX_PAUSE:
			val |= FCR1_FDFCRX_EN;
			break;

		case FLOW_CONTROL_TX_PAUSE:
			val |= FCR1_FDFCTX_EN;
			break;

		case FLOW_CONTROL_SYMMETRIC:
			val |= FCR1_FDFCTX_EN | FCR1_FDFCRX_EN;
			break;
#ifdef NEVER
		case FLOW_CONTROL_NONE:
			if (!dp->full_duplex) {
				val |=  FCR1_HDFC_EN;
			}
			break;
#endif
		}
		OUTB(dp, FCR1, val);
	}
}

static void
rh_get_stats(struct gem_dev *dp)
{
	volatile int	x;
	struct rh_dev	*lp = (struct rh_dev *)dp->private;

	if (IS_RHINE2_OR_LATER(lp->revid)) {
		dp->stats.missed += INW(dp, MPAC);
	}
	else {
		/* they are unreliable on rhine I */
		x = INW(dp, MPAC);
	}
	x = INW(dp, CRCC);
	OUTW(dp, MPAC, 0);
}

/*
 * discriptor manupiration
 */
#ifdef DEBUG_LEVEL
static int rh_tx_frags[GEM_MAXTXFRAGS];
#endif
static int
rh_tx_desc_write(struct gem_dev *dp, uint_t slot,
		ddi_dma_cookie_t *dmacookie, int frags, uint32_t flags)
{
	int			i;
	struct tx_desc		*tdp;
	ddi_dma_cookie_t	*dcp;
	uint32_t		mark;
	struct rh_dev		*lp = (struct rh_dev *)dp->private;
	ddi_acc_handle_t	h = dp->desc_acc_handle;

#if DEBUG_LEVEL > 2
	cmn_err(CE_CONT,
	"!%s: rh_tx_desc_write time: %d, seqnum: %d, slot %d, frags: %d flags: %d",
		dp->name, ddi_get_lbolt(), dp->tx_desc_tail, slot, frags, flags);
	for (i = 0; i < frags; i++) {
		cmn_err(CE_CONT, "!%d: addr: 0x%llx, len: 0x%x",
			i, dmacookie[i].dmac_laddress, dmacookie[i].dmac_size);
	}
#endif
#if DEBUG_LEVEL > 3
	flags |= GEM_TXFLAG_INTR;
#endif
#ifdef DEBUG_LEVEL
	rh_tx_frags[min(frags, GEM_MAXTXFRAGS) - 1]++;
#endif
	/*
	 * To avoid condition race, enabme PTx interrupt before making
	 * descriptor(s) for the packet
	 */
#ifdef OPT_TX_INTR
	if ((dp->misc_flag & GEM_NOINTR) == 0 && IS_RHINE1(lp->revid) &&
	    (flags & GEM_TXFLAG_INTR) != 0 && (lp->imr & ISR_PTX) == 0) {
		/*
		 * enable PTX interrupt for the time being
		 */
		lp->imr |= ISR_PTX;
#if 0
		if (mutex_tryenter(&dp->intrlock)) {
			if (lp->imr_hw != 0) {
				lp->imr_hw = lp->imr;
				RH_WRITE_IMR(dp, lp->imr_hw);
				mutex_exit(&dp->intrlock);
			}
		} else
#endif
		{
			lp->imr_change_req = TRUE;
			ddi_trigger_softintr(lp->soft_id);
		}
	}
#endif /* OPT_TX_INTR */
	/*
	 * write tx descriptor(s) in reversed order
	 */
	mark = TDES1_EDP | TDES1_C | ((flags & GEM_TXFLAG_INTR) ? TDES1_IC : 0);
	for (i = frags - 1, dcp = &dmacookie[frags - 1]; i > 0; i--, dcp--) {

		tdp = &((struct tx_desc *)dp->tx_ring)[
				SLOT(slot + i, TX_RING_SIZE)];
		ddi_put32(h, &tdp->td_length, dcp->dmac_size | mark);
		ddi_put32(h, &tdp->td_addr, dcp->dmac_address);
		ddi_put32(h, &tdp->td_csr, 0);
		mark = TDES1_C;

		ddi_dma_sync(dp->desc_dma_handle,
			(off_t)(((caddr_t)tdp) - dp->rx_ring),
			sizeof(struct tx_desc), DDI_DMA_SYNC_FORDEV);
	}

	/* for the first fragment */
	tdp = &((struct tx_desc *)dp->tx_ring)[slot];
	ddi_put32(h, &tdp->td_length, dcp->dmac_size | mark | TDES1_STP);
	ddi_put32(h, &tdp->td_addr, dcp->dmac_address);

	if ((flags & GEM_TXFLAG_INTR) == 0 && IS_RHINE2_OR_LATER(lp->revid)) {
		/* set suppress-interrupt request for the packet */
		ddi_put32(h, &tdp->td_next,
			ddi_get32(h, &tdp->td_next) | TDES3_TDCTL);
	}
	ddi_put32(h, &tdp->td_csr, TDES0_OWN);

	ddi_dma_sync(dp->desc_dma_handle,
		(off_t)(((caddr_t)tdp) - dp->rx_ring),
		sizeof(struct tx_desc), DDI_DMA_SYNC_FORDEV);

	/* kick Tx engine */
#ifdef TXTIMEOUT_TEST
	if ((rh_send_cnt++ % 100) == 99) {
		lp->cr &= ~CR_TXON;
		OUTW(dp, CR, lp->cr);
	}
#endif
	if (IS_RHINE3M_OR_LATER(lp->revid)) {
		OUTB(dp, TXQW, 1 << 7);
	}
	else {
		RH_KICK_TX(dp, lp->cr);
	}

	return frags;
}

static int
rh_rx_desc_write(struct gem_dev *dp, uint_t slot,
	    ddi_dma_cookie_t *dmacookie, int frags)
{
	struct rx_desc		*rdp;
	int			i;
	struct rh_dev		*lp = (struct rh_dev *)dp->private;
	ddi_acc_handle_t	h = dp->desc_acc_handle;

#if DEBUG_LEVEL > 2
	cmn_err(CE_CONT,
		"!%s: rh_rx_desc_write seqnum: %d, slot %d, frags: %d",
		dp->name, dp->rx_desc_tail, slot, frags);
	for (i = 0; i < frags; i++) {
		cmn_err(CE_CONT, "!%d: addr: 0x%llx, len: 0x%x",
			i, dmacookie[i].dmac_laddress, dmacookie[i].dmac_size);
	}
#endif
	/*
	 * write a RX descriptor
	 */
	rdp = &((struct rx_desc *)dp->rx_ring)[slot];

	ddi_put32(h, &rdp->rd_addr, dmacookie->dmac_address);
	ddi_put32(h, &rdp->rd_length, dmacookie->dmac_size);
	ddi_put32(h, &rdp->rd_status, RDES0_OWN);

	ddi_dma_sync(dp->desc_dma_handle,
		(off_t)(((caddr_t)rdp) - dp->rx_ring),
		sizeof(struct rx_desc), DDI_DMA_SYNC_FORDEV);

	if (IS_RHINE3_OR_LATER(lp->revid) && dp->nic_active) {
		/* Tx flow control for rhine III */
		OUTB(dp, FCR0, 1);
	}

	return frags;
}

static void
rh_tx_desc_dump(struct gem_dev *dp, seqnum_t head, int ndesc)
{
	int			i;
	struct tx_desc		*tdp;
	struct rh_dev		*lp = (struct rh_dev *)dp->private;
	ddi_acc_handle_t	h = dp->desc_acc_handle;

	cmn_err(CE_CONT, "!%s: tx_desc_dump:  slot:%d, seqnum:%d",
		dp->name,SLOT(head, TX_RING_SIZE), head);

	for (i = 0; i < ndesc; i++) {
		tdp = &((struct tx_desc *)dp->tx_ring)[
				SLOT(head + i, TX_RING_SIZE)];

		ddi_dma_sync(dp->desc_dma_handle,
			(off_t)(((caddr_t)tdp) - dp->rx_ring),
			sizeof(struct tx_desc), DDI_DMA_SYNC_FORKERNEL);

		cmn_err(CE_CONT, "! %d: %b, %b, 0x%08x",
			i, ddi_get32(h, &tdp->td_csr), TSR_BITS,
			ddi_get32(h, &tdp->td_length), TDES1_BITS,
			ddi_get32(h, &tdp->td_addr));
	}
}

static uint_t
rh_tx_desc_stat(struct gem_dev *dp, int slot, int ndesc)
{
	struct tx_desc		*tdp;
	uint32_t		csr;
	uint_t			ret;
	int			cols;
	int			i;
	struct rh_dev		*lp = (struct rh_dev *)dp->private;
	ddi_acc_handle_t	h = dp->desc_acc_handle;

	/*
	 * check first descriptor of the packet
	 */
	tdp = &((struct tx_desc *)dp->tx_ring)[slot];

	ddi_dma_sync(dp->desc_dma_handle,
		(off_t)(((caddr_t)tdp) - dp->rx_ring),
		sizeof(struct tx_desc), DDI_DMA_SYNC_FORKERNEL);

	csr = ddi_get32(h, &tdp->td_csr);

#if DEBUG_LEVEL > 2
	if (ndesc > 0) {
		int		i;
		uint32_t	isr;

		RH_READ_ISR(dp, isr);
		cmn_err(CE_CONT,
			"!%s: rh_tx_desc_stat: time: %d "
			"slot:%d ndesc:%d isr:%b misr:%b",
			dp->name, ddi_get_lbolt(), slot, ndesc,
			isr & 0xffff, ISR_BITS, isr >> 16, MISR_BITS);
		for (i = 0; i < ndesc; i++) {
			struct tx_desc	*tp;

			tp = &((struct tx_desc *)dp->tx_ring)[
						SLOT(slot + i, TX_RING_SIZE)];
			cmn_err(CE_CONT, "!%d: tsr: %b txdesc1: %b",
			    i,
			    ddi_get32(h, &tp->td_csr), TSR_BITS,
			    ddi_get32(h, &tp->td_length), TDES1_BITS);
		}
	}
#endif
	if ((csr & TDES0_OWN) != 0) {
		/* not transmitted yet */
		return 0;
	}

	ret = GEM_TX_DONE;

	if (lp->td_unreliable) {
		goto x;
	}

#define	TSR_ERRS	\
	(TSR_SERR | TSR_TBUFF | TSR_UDF | TSR_CRS | TSR_OWC | TSR_ABT | TSR_CDH)
	if ((csr & TSR_TXERR) == 0) {
		if (!dp->full_duplex && (cols = (csr & TSR_NCR)) > 0) {
			dp->stats.collisions += cols;
			if (cols == 1) {
				dp->stats.first_coll++;
			}
			else /* if (cols > 1)*/ {
				dp->stats.multi_coll++;
			}
		}
		/* short cut for normal case */
		goto x;
	}

	DPRINTF(2, (CE_CONT, "!%s: tx error: tsr: %b  desc:%d",
		dp->name, csr, TSR_BITS, dp->tx_desc_head));

	dp->stats.errxmt++;

	if ((csr & TSR_CRS) != 0) {
		dp->stats.nocarrier++;
	}
	if ((csr & TSR_OWC) != 0) {
		dp->stats.xmtlatecoll++;
	}
	if ((csr & TSR_UDF) != 0) {
		dp->stats.underflow++;

		lp->tft = min(lp->tft+1, RTFT_SF);

		OUTB(dp, TCR,
			(lp->tft << TCR_RTFT_SHIFT) |
			TCR_LB_NORMAL | TCR_OFSET);

		OUTW(dp, BCR,
			(lp->tft << BCR_CTFT_SHIFT) |
			(lp->rft << BCR_CRFT_SHIFT) |
			lp->maxdma);

		/* transmitter halted, resend the packet */
		ndesc = 0;
		ret |= GEM_TX_ERR;
	}
	if ((csr & TSR_ABT) != 0) {
		if (!dp->full_duplex) {
			dp->stats.excoll++;
			dp->stats.collisions += 16;
		}
#ifdef notdef
		rh_tx_desc_dump(dp, dp->tx_desc_head, ndesc);
#endif
		/* transmitter halted, drop the packet */
		ret |= GEM_TX_ERR;
	}

	if ((ret & GEM_TX_ERR) == 0) {
		/* no need to recover from a fatal error */
		goto x;
	}

	/*
	 * Try to recover the transmitter state without resetting the chip.
	 */
	/* wait until the transmitter becomes idle */
	i = 0;
	while ((INW(dp, CR) & CR_TXON) != 0) {
		if (i++ > 10) {
			DPRINTF(2, (CE_WARN,
				"!%s: rh_tx_desc_stat: timeout tx stop",
				dp->name));
			goto x;
		}
		drv_usecwait(10);
	}
	DPRINTF(2, (CE_CONT, "!%s: tx stopped in %d uS", dp->name, i*10));

	/* fix Current Tx Desctriptor Address */
	OUTL(dp, CTDA, ((uint32_t)dp->tx_ring_dma) +
		    SLOT(slot + ndesc, TX_RING_SIZE) * sizeof(struct tx_desc));

	/* fix the next actions to do. */
	ret = (ndesc > 0) ? GEM_TX_DONE : 0;

	/* enable the transmitter again. */
	OUTW(dp, CR, lp->cr);

	/* kick tx polling */
	RH_KICK_TX(dp, lp->cr);
x:
	/* reset the supress-intr bit */
	if (ret != 0) {
		/*
		 * As tx was finished, update the interrupt mask.
		 */
		if (IS_RHINE1(lp->revid)) {
#ifdef OPT_TX_INTR
			if (dp->tx_desc_intr == dp->tx_desc_head + ndesc &&
			   (lp->imr & ISR_PTX) != 0) {
				/*
				 * XXX - Hardware IMR will be updated later.
				 */
				lp->imr &= ~ISR_PTX;
			}
#endif
		}
		else {
			/*
			 * For Rhine II or later:
			 * Ensure that there is no suppress interrupt request
			 * in TD3.
			 */
			ddi_put32(h, &tdp->td_next,
				ddi_get32(h, &tdp->td_next) & TDES3_NEXT);
		}
	}
	return ret;
}

static uint_t
rh_rx_desc_stat(struct gem_dev *dp, int slot, int ndesc)
{
	struct rx_desc		*rdp;
	uint32_t		rsr;
	uint_t			ret;
	uint_t			len;
	struct rh_dev		*lp = (struct rh_dev *)dp->private;
	ddi_acc_handle_t	h = dp->desc_acc_handle;

	rdp = &((struct rx_desc *)dp->rx_ring)[slot];

	ddi_dma_sync(dp->desc_dma_handle,
		(off_t)(((caddr_t)rdp) - dp->rx_ring),
		sizeof(struct rx_desc), DDI_DMA_SYNC_FORKERNEL);

	rsr = ddi_get32(h, &rdp->rd_status);
	len = (rsr >> RDES0_LEN_SHIFT) & RDES0_LEN_MASK;
	ret = GEM_RX_DONE;

	DPRINTF(2, (CE_CONT,
		"!%s: rh_rx_desc_stat: slot:%d dma_addr:0x%x len:%d rsr 0x%b",
		dp->name, slot,
		ddi_get32(h, &rdp->rd_addr),
		ddi_get32(h, &rdp->rd_length),
		ddi_get32(h, &rdp->rd_status), RSR_BITS));

	if ((rsr & RDES0_OWN) != 0) {
		/* it isn't received yet */
		return 0;
	}

	if ((rsr & (RSR_STP | RSR_EDP)) != (RSR_STP | RSR_EDP)) {
		/* big packet, ignore this fragment */
		if ((rsr & RSR_STP) != 0) {
			dp->stats.errrcv++;
			dp->stats.frame_too_long++;
		}
		return GEM_RX_DONE | GEM_RX_ERR;
	}

	if ((rsr & RSR_RXOK) == 0) {
		/* errored packet */
		dp->stats.errrcv++;
		if ((rsr & RSR_LONG) != 0) {
			dp->stats.frame_too_long++;
		}
		if ((rsr & RSR_RUNT) != 0) {
			dp->stats.runt++;
			DPRINTF(2, (CE_CONT,
		"!%s: rh_rx_desc_stat: slot:%d dma_addr:0x%x len:%d rsr 0x%b",
				dp->name, slot,
				ddi_get32(h, &rdp->rd_addr),
				ddi_get32(h, &rdp->rd_length),
				ddi_get32(h, &rdp->rd_status), RSR_BITS));
		}
		if ((rsr & (RSR_SERR | RSR_BUFF)) != 0) {
			dp->stats.overflow++;
		}
		if ((rsr & RSR_FAE) != 0) {
			dp->stats.frame++;
		}
		if ((rsr & RSR_CRC) != 0) {
			dp->stats.crc++;
		}

		ret |= GEM_RX_ERR;
	}

	if (len >= ETHERFCSL) {
		len -= ETHERFCSL;
	}

	return ret | (len & GEM_RX_LEN);
}

static void
rh_tx_desc_init(struct gem_dev *dp, int slot)
{
	struct tx_desc		*tdp;
	ddi_acc_handle_t	h = dp->desc_acc_handle;

	tdp = &((struct tx_desc *)dp->tx_ring)[slot];

	/* invalidate this descriptor */
	ddi_put32(h, &tdp->td_csr, 0);

	ddi_dma_sync(dp->desc_dma_handle,
		(off_t)(((caddr_t)tdp) - dp->rx_ring),
		sizeof(struct tx_desc), DDI_DMA_SYNC_FORDEV);

	/* link to previous descriptor */
	tdp = &((struct tx_desc *)dp->tx_ring)[SLOT(slot - 1, TX_RING_SIZE)];
	ddi_put32(h, &tdp->td_next,
		((uint32_t)dp->tx_ring_dma) + slot*sizeof(struct tx_desc));

	ddi_dma_sync(dp->desc_dma_handle,
		(off_t)(((caddr_t)tdp) - dp->rx_ring),
		sizeof(struct tx_desc), DDI_DMA_SYNC_FORDEV);
}

static void
rh_rx_desc_init(struct gem_dev *dp, int slot)
{
	struct rx_desc		*rdp;
	ddi_acc_handle_t	h = dp->desc_acc_handle;

	rdp = &((struct rx_desc *)dp->rx_ring)[slot];

	/* invalidate this descriptor */
	ddi_put32(h, &rdp->rd_status, 0);

	ddi_dma_sync(dp->desc_dma_handle,
		(off_t)(((caddr_t)rdp) - dp->rx_ring),
		sizeof(struct rx_desc), DDI_DMA_SYNC_FORDEV);

	/* link this to its previous descriptor */
	rdp = &((struct rx_desc *)dp->rx_ring)[SLOT(slot - 1, RX_RING_SIZE)];
	ddi_put32(h, &rdp->rd_next,
		((uint32_t)dp->rx_ring_dma) + slot*sizeof(struct rx_desc));

	ddi_dma_sync(dp->desc_dma_handle,
		(off_t)(((caddr_t)rdp) - dp->rx_ring),
		sizeof(struct rx_desc), DDI_DMA_SYNC_FORDEV);
}

static void
rh_tx_desc_clean(struct gem_dev *dp, int slot)
{
	struct tx_desc		*tdp;
	ddi_acc_handle_t	h = dp->desc_acc_handle;

	tdp = &((struct tx_desc *)dp->tx_ring)[slot];

	/* invalidate this descriptor */
	ddi_put32(h, &tdp->td_csr, 0);
	ddi_put32(h, &tdp->td_next, ddi_get32(h, &tdp->td_next) & TDES3_NEXT);

	ddi_dma_sync(dp->desc_dma_handle,
		(off_t)(((caddr_t)tdp) - dp->rx_ring),
		sizeof(struct tx_desc), DDI_DMA_SYNC_FORDEV);
}

static void
rh_rx_desc_clean(struct gem_dev *dp, int slot)
{
	struct rx_desc		*rdp;
	ddi_acc_handle_t	h = dp->desc_acc_handle;

	rdp = &((struct rx_desc *)dp->rx_ring)[slot];

	/* invalidate this descriptor */
	ddi_put32(h, &rdp->rd_status, 0);

	ddi_dma_sync(dp->desc_dma_handle,
		(off_t)(((caddr_t)rdp) - dp->rx_ring),
		sizeof(struct rx_desc), DDI_DMA_SYNC_FORDEV);
}

/*
 * Device depend interrupt handler
 */
#ifdef OPT_TX_INTR
static
uint_t rh_update_imr(struct gem_dev *dp)
{
	uint_t		ret;
	struct rh_dev	*lp = (struct rh_dev *)dp->private;

	ret = DDI_INTR_UNCLAIMED;

	mutex_enter(&dp->intrlock);
	mutex_enter(&dp->xmitlock);

	if (lp->imr_change_req) {
		if ((dp->misc_flag & GEM_NOINTR) == 0 && lp->imr_hw != 0 &&
		     lp->imr_hw != lp->imr) {
			lp->imr_hw = lp->imr;
			RH_WRITE_IMR(dp, lp->imr_hw);
		}
		lp->imr_change_req = FALSE;
		ret = DDI_INTR_CLAIMED;
	}

	mutex_exit(&dp->xmitlock);
	mutex_exit(&dp->intrlock);

	return ret;
}
#endif /* OPT_TX_INTR */

static u_int
rh_interrupt(struct gem_dev *dp)
{
	uint32_t	isr;
	uint_t		flag = 0;
	boolean_t	need_to_reset = FALSE;
	struct rh_dev	*lp = (struct rh_dev *)dp->private;

	RH_READ_ISR(dp, isr);

	if ((isr & lp->imr_hw) == 0) {
		/* Not for us */
		return DDI_INTR_UNCLAIMED;
	}

	DPRINTF(2, (CE_CONT, "!%s: Interrupt, isr: %b, misr: %b",
		dp->name, isr & 0xffff, ISR_BITS, isr >> 16, MISR_BITS));

	/* disable interrupt */
	RH_WRITE_IMR(dp, 0);

	if (!dp->nic_active) {
		/* the device is not active, no more interrupts */
		lp->imr = 0;
		lp->imr_hw = lp->imr;
		RH_WRITE_ISR(dp, isr);

		return DDI_INTR_CLAIMED;
	}
#ifdef RESET_TEST
	lp->reset_test++;
	if ((lp->reset_test % 10000) == 9999) {
		need_to_reset = TRUE;
	}
#endif

	/* clear interrupts */
	RH_WRITE_ISR(dp, isr); FLUSH(dp);

	isr &= lp->imr_hw;

	if ((isr & ISR_CNT) != 0) {
		/* statics counter overflow */
		rh_get_stats(dp);
	}

	if (isr & (ISR_PRX | ISR_RXE | ISR_PKRACE | ISR_RU | ISR_NORBF)) {

		/* packet was received, or receive error happened */
		/*
		 * RU   : receive buffer unavailable
		 */
		gem_receive(dp);
	}

	if ((isr & ISR_OVFI) != 0) {
		/*
		 * receiving FIFO overflow
		 */
		/* Do nothing */
	}

	if ((isr & ISR_BE) != 0) {
		cmn_err(CE_WARN, "%s: unexpected interrupt: isr:%b misr:%b",
			dp->name, isr & 0xffff, ISR_BITS, isr >> 16, MISR_BITS);
		need_to_reset = TRUE;
		goto x;
	}

	if ((isr & (MISR_TDWBI << 16)) != 0) {
		int	cnt;

		DPRINTF(2, (CE_WARN, "!%s: TDWBI interrupt: isr:%b misr:%b",
			dp->name, isr & 0xffff, ISR_BITS,
			isr >> 16, MISR_BITS));

		mutex_enter(&dp->xmitlock);

		/* XXX - stop Tx */
		cnt = 0;
		while ((INW(dp, CR) & CR_TXON) != 0) {
			cnt++;
			if (cnt == 10) {
				break;
			}
			drv_usecwait(10);
		}

		DPRINTF(2, (CE_CONT, "!%s: tx stopped on TDWBI in %d uS",
			dp->name, cnt*10));

		if (cnt == 10) {
			/* timeout: we failed to stop tx side */
			need_to_reset = TRUE;
			mutex_exit(&dp->xmitlock);
			goto x;
		}

		lp->td_unreliable = TRUE;
		(void)gem_reclaim_txbuf(dp);
		lp->td_unreliable = FALSE;

		/*
		 * Fix Current Tx Desctriptor to point
		 * the head of tx list.
		 */
		OUTL(dp, CTDA, ((uint32_t)dp->tx_ring_dma) +
			    SLOT(dp->tx_desc_head, TX_RING_SIZE)
				* sizeof(struct tx_desc));

		/* XXX - start Tx */
		OUTW(dp, CR, lp->cr);
		RH_KICK_TX(dp, lp->cr);

		mutex_exit(&dp->xmitlock);
	}

	if ((isr & ISR_PTX) != 0) {
		if (gem_tx_done(dp)) {
			flag |= INTR_RESTART_TX;
		}
	}

	if ((isr & (ISR_TXE | ISR_TU | ISR_ABTI)) != 0) {
		/*
		 * Need to process transmitted buffer immediately.
		 */
		if (gem_tx_done(dp)) {
			flag |= INTR_RESTART_TX;
		}
	}
#ifdef notdef
	if ((isr & ISR_SRCI) != 0) {
		/*
		 * PHY port status changed.
		 */
		DPRINTF(0, (CE_WARN,
			"%s: port status changed: isr:%b misr:%b",
			dp->name,
			isr & 0xffff, ISR_BITS, isr >> 16, MISR_BITS));
		(void) gem_mii_link_check(dp);
	}
#endif
	if ((isr & (ISR_TU | ISR_ETI/*| ISR_KEYI*/)) != 0) {
		/*
		 * unexpected interrupt
		 */
		/* BE: PCI bus error */
		/* TU: Transmit buffer underflow */
		/* ETI: Transmit descriptor underflow */
		/* KEYI : magic packet received / gpio  */

		cmn_err(CE_WARN,
			"%s: unexpected interrupt: isr:%b misr:%b",
			dp->name,
			isr & 0xffff, ISR_BITS, isr >> 16, MISR_BITS);
	}
x:
	RH_READ_ISR(dp, isr);

	if (need_to_reset) {
		mutex_enter(&dp->xmitlock);
		gem_restart_tx(dp);
		mutex_exit(&dp->xmitlock);
		flag |= INTR_RESTART_TX;
	}

	if ((dp->misc_flag & GEM_NOINTR) == 0) {
		/* enable interrupts again */
		if (lp->imr_hw != lp->imr) {
			mutex_enter(&dp->xmitlock);
			if (lp->imr_hw != lp->imr) {
				lp->imr_hw = lp->imr;
			}
			mutex_exit(&dp->xmitlock);
		}
		RH_WRITE_IMR(dp, lp->imr_hw);
	}

	return DDI_INTR_CLAIMED | flag;
}

/*
 * HW depend MII routines
 */
static void
rh_mii_sync(struct gem_dev *dp)
{
	/* nothing to do */
}

static uint16_t
rh_mii_read(struct gem_dev *dp, int reg)
{
	int	n;

	OUTB(dp, MIIADR, reg);
	OUTB(dp, MIICR, INB(dp, MIICR) | MIICR_RCMD);

	n = 100;
	while ((INB(dp, MIICR) & MIICR_RCMD) != 0) {
		if  (--n == 0) {
			cmn_err(CE_WARN, "%s: mii_read: timeout", dp->name);
			return 0xffff;
		}
		drv_usecwait(10);
	}
	return INW(dp, MIIDATA);
}

static void
rh_mii_write_raw(struct gem_dev *dp, int reg, uint16_t val)
{
	int	n;

	OUTB(dp, MIIADR, reg);
	OUTW(dp, MIIDATA, val);
	OUTB(dp, MIICR, INB(dp, MIICR) | MIICR_WCMD);
	n = 100;
	while ((INB(dp, MIICR) & MIICR_WCMD) != 0) {
		if (--n == 0) {
			cmn_err(CE_WARN, "%s: mii_write: timeout", dp->name);
			return;
		}
		drv_usecwait(10);
	}
}

static void
rh_mii_write(struct gem_dev *dp, int reg, uint16_t val)
{
	uint16_t	ret;
	struct rh_dev	*lp = (struct rh_dev *)dp->private;

	if (reg == MII_CONTROL &&
	   (IS_RHINE3_OR_LATER(lp->revid) ||
		(IS_RHINE2_OR_LATER(lp->revid) && IS_VT6103(dp->mii_phy_id)))) {

		ret = rh_mii_read(dp, 0x10);
		if ((val & MII_CONTROL_ANE) != 0) {
			ret &= ~1;
		} else {
			ret |= 1;
		}
		rh_mii_write_raw(dp, 0x10, ret);
	}

	rh_mii_write_raw(dp, reg, val);
}

/* ======================================================== */
/*
 * OS depend (device driver) routine
 */
/* ======================================================== */
static int
rhattach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int			i;
	int			n;
	ddi_iblock_cookie_t	c;
	ddi_acc_handle_t	conf_handle;
	int			ret;
	uint16_t		vid;
	uint16_t		did;
	uint8_t			revid;
	int			unit;
	struct chip_info	*p;
	u_int			len;
	struct pci_phys_spec	*regs;
	const char		*drv_name;
	struct gem_dev		*dp;
	struct rh_dev	*lp;
	int			rno_io;
	int			rno_mm;
	void			*base_mm;
	void			*base_io;
	ddi_acc_handle_t	ha_mm;
	ddi_acc_handle_t	ha_io;
	struct gem_conf		*gcp;
	uint8_t			stick;
	uint32_t		cap_ptr;
	uint32_t		cap;
	uint32_t		ps;
	uint32_t		ilr;
	uint8_t			reg;

	unit =  ddi_get_instance(dip);
	drv_name = ddi_driver_name(dip);

	DPRINTF(2, (CE_CONT, "!%s%d: rhattach: called at time:%d",
		drv_name, unit, ddi_get_lbolt()));

	if (cmd == DDI_ATTACH) {
		/*
		 * Check if chip is supported.
		 */
		if (pci_config_setup(dip, &conf_handle) != DDI_SUCCESS) {
			cmn_err(CE_WARN, "%s%d: ddi_regs_map_setup failed",
				drv_name, unit);
			return DDI_FAILURE;
		}
		vid  = pci_config_get16(conf_handle, PCI_CONF_VENID);
		did  = pci_config_get16(conf_handle, PCI_CONF_DEVID);
		revid= pci_config_get8(conf_handle, PCI_CONF_REVID);

		for (i = 0, p = rh_chiptbl; i < CHIPTABLESIZE; i++, p++) {
			if (p->venid == vid && p->devid == did) {
				/* found */
				cmn_err(CE_CONT,
			"!%s%d: %s (vid: 0x%04x, did: 0x%04x, revid: 0x%02x)",
				drv_name, unit, p->name, vid, did, revid);
				goto chip_found;
			}
		}
		/* Not found */
		cmn_err(CE_WARN,
			"%s: rh_attach: wrong PCI venid/devid (0x%x, 0x%x)",
			drv_name, vid, did);
		goto err;

chip_found:
		/* ensure the pmr status is  D0 mode */
		cap_ptr = pci_config_get8(conf_handle, PCI_CONF_CAP_PTR);
		while (cap_ptr != 0) {
			/* read capability header */
			cap = pci_config_get32(conf_handle, cap_ptr);
			if ((cap & 0xff) == PCI_CAP_ID_PM) {
				/*
				 * A power management capability found.
				 * ensure D0 mode
				 */
				ps = pci_config_get32(conf_handle, cap_ptr + 4);

				DPRINTF(0, (CE_CONT,
				"!%s%d: pmc found at 0x%x: ps: 0x%08x",
				drv_name, unit, cap_ptr, ps));

				ps &= ~PCI_PMCSR_STATE_MASK;
				pci_config_put32(conf_handle, cap_ptr + 4, ps);

				/* no need to work more */
				break;
			}
			/* get next_ptr */
			cap_ptr = (cap >> 8) & 0xff;
		}

		/* Undocumented dances and songs */
		if (IS_RHINE2_OR_LATER(revid)) {
			pci_config_put8(conf_handle, 0x53, 
				pci_config_get8(conf_handle, 0x53) | 0x04);
		}
		if (IS_RHINE3LOM(revid)) {
			pci_config_put8(conf_handle, 0x52,
				pci_config_get8(conf_handle, 0x52) | 0x02);
		}

		ilr = pci_config_get32(conf_handle, PCI_CONF_ILINE);
		DPRINTF(0, (CE_CONT, "!%s%d: ilr 0x%08x", drv_name, unit, ilr));

		/*
		 * Map the device registers in the kernel virtual space.
		 */

		/* Search IO-range or memory-range to be mapped */
		regs = NULL;
		len  = 0;

		if (ddi_prop_lookup_int_array(
			DDI_DEV_T_ANY, dip, DDI_PROP_DONTPASS,
			"reg", (int **)&regs, &len) != DDI_PROP_SUCCESS) {

			cmn_err(CE_WARN,
			"%s%d: failed to get reg property", drv_name, unit);
			goto err;
		}
		n = len / (sizeof(struct pci_phys_spec) / sizeof(int));

		ASSERT(regs != NULL && len > 0);
#if DEBUG_LEVEL > 2
		for (i = 0; i < n; i++) {
			cmn_err(CE_CONT,
				"!%s%d: regs[%d]: %08x.%08x.%08x.%08x.%08x",
				drv_name, unit, i,
				regs[i].pci_phys_hi,
				regs[i].pci_phys_mid,
				regs[i].pci_phys_low,
				regs[i].pci_size_hi,
				regs[i].pci_size_low);
		}
#endif
		rno_io = 0;
		rno_mm = 0;
		for (i = 1; i < n; i++) {
			if ((regs[i].pci_phys_hi & PCI_REG_ADDR_M) == 
						PCI_ADDR_IO && rno_io == 0) {
					rno_io = i;
			}
			if ((regs[i].pci_phys_hi & PCI_REG_ADDR_M) == 
#ifdef MAP_MEM
						PCI_ADDR_MEM32
#else
						PCI_ADDR_IO
#endif
						&& rno_mm == 0
			) {
					rno_mm = i;
			}
		}
		ddi_prop_free(regs);

		if (rno_mm == 0 || rno_io == 0) {
			cmn_err(CE_WARN, "%s%d: failed to find IO space",
				drv_name, unit);
			goto err;
		}

		/* ensure we can access the registers through IO space. */
		ret = pci_config_get16(conf_handle, PCI_CONF_COMM);
		ret |= PCI_COMM_IO | PCI_COMM_ME | PCI_COMM_MAE;
		pci_config_put16(conf_handle, PCI_CONF_COMM, ret);

		if (ddi_regs_map_setup(dip, rno_io, (caddr_t *)&base_io,
			0, 0, &rh_dev_attr, &ha_io) != DDI_SUCCESS) {
			cmn_err(CE_WARN, "%s%d: ddi_regs_map_setup(io) failed",
				drv_name, unit);
			goto err;
		}

		if (ddi_regs_map_setup(dip, rno_mm, (caddr_t *)&base_mm,
			0, 0, &rh_dev_attr, &ha_mm) != DDI_SUCCESS) {
			cmn_err(CE_WARN, "%s%d: ddi_regs_map_setup(mem) failed",
				drv_name, unit);
			ddi_regs_map_free(&ha_io);
			goto err;
		}

		/* disable WOL */
		if (IS_RHINE2_OR_LATER(revid)) {
			stick = ddi_get8(ha_io, ((uint8_t *)base_io)+STICKHW);
			ddi_put8(ha_io, ((uint8_t *)base_io)+STICKHW,
					stick & ~(STICKHW_DS1 | STICKHW_DS0));
			ddi_put8(ha_io, ((uint8_t *)base_io)+WOLCR_CLR,
					WOLCR_LinkOFF);
			ddi_put8(ha_io, ((uint8_t *)base_io)+WOLCG_CLR, 0xFF);
			ddi_put8(ha_io, ((uint8_t *)base_io)+PWRCSR_CLR, 0xFF);
		}

		/* we use eeprom autoload function to get mac address */
#ifdef DEBUG_LEVEL
		/* write a dummy to ensure eeprom autoloading works. */
		for (i = 0; i < ETHERADDRL; i++) {
			ddi_put8(ha_io, ((uint8_t *)base_io)+PAR+i, i);
		}
#endif
		ddi_get8(ha_io, ((uint8_t *)base_io)+EECSR);
		ddi_put8(ha_io, ((uint8_t *)base_io)+EECSR, reg|EECSR_AUTOLD);

		/*
		 * XXX - Don't access rhine for 10mS.
		 * Rhine doesn't respond to any PCI transactions
		 * while it is loading eeprom. On sparc, it causes
		 * a bus error panic. We must for wait 2.5 mS at least.
		 */
		drv_usecwait(10000);

		/* enable memory mapped I/O */
		/* XXX - reloading eeprom will disable MMIO */
		if (IS_RHINE1_VT86C100A(revid)) {
			reg = ddi_get8(ha_io, ((uint8_t *)base_io)+CFGA);
			ddi_put8(ha_io, ((uint8_t *)base_io)+CFGA,
				reg | CFGA_MMIOEN);
		}
		else {
			reg = ddi_get8(ha_io, ((uint8_t *)base_io)+CFGD);
			ddi_put8(ha_io, ((uint8_t *)base_io)+CFGD,
				reg | CFGD_MMIOEN);
		}                                                        
		ddi_regs_map_free(&ha_io);

		/*
		 * construct gem configration
		 */
		gcp = (struct gem_conf *) kmem_zalloc(sizeof(*gcp), KM_SLEEP);

		/* name */
		sprintf(gcp->gc_name, "%s%d", drv_name, unit);

		/* consistency on tx and rx */
		gcp->gc_tx_buf_align = 
			IS_RHINE2_OR_LATER(revid)
			? (sizeof(uint8_t) - 1)
			: (sizeof(uint32_t) - 1);
#ifdef PTX_TEST
		gcp->gc_tx_max_frags = 1;
#else
		/* rhine I supports only a fragment per tx packet */
		gcp->gc_tx_max_frags = 
			IS_RHINE2_OR_LATER(revid) ? GEM_MAXTXFRAGS : 1;
#endif
		gcp->gc_tx_desc_size =
			IS_RHINE3M_OR_LATER(revid)
			? sizeof(struct tx_desc) * TX_RING_SIZE * 8
			: sizeof(struct tx_desc) * TX_RING_SIZE;

		gcp->gc_tx_ring_size = TX_RING_SIZE;
		gcp->gc_tx_buf_size  = TX_BUF_SIZE;
		gcp->gc_tx_max_descs_per_pkt = gcp->gc_tx_max_frags;
		gcp->gc_tx_auto_pad  = FALSE;
		gcp->gc_tx_copy_thresh = rh_tx_copy_thresh;

		gcp->gc_rx_buf_align = sizeof(uint32_t) - 1;
		gcp->gc_rx_max_frags = 1;
		gcp->gc_rx_desc_size = sizeof(struct rx_desc) * RX_RING_SIZE;
		gcp->gc_rx_ring_size = RX_RING_SIZE;
		gcp->gc_rx_buf_size  = RX_BUF_SIZE;
		gcp->gc_rx_max_descs_per_pkt = gcp->gc_rx_max_frags;
		gcp->gc_rx_copy_thresh = rh_rx_copy_thresh;
		gcp->gc_rx_buf_max   = gcp->gc_rx_buf_size + 1;

		/* map attributes */
		STRUCT_COPY(gcp->gc_dev_attr, rh_dev_attr);
		STRUCT_COPY(gcp->gc_buf_attr, rh_buf_attr);
		STRUCT_COPY(gcp->gc_desc_attr, rh_dev_attr);

		/* dma attributes */
		STRUCT_COPY(gcp->gc_dma_attr_desc, rh_dma_attr_desc);
		STRUCT_COPY(gcp->gc_dma_attr_txbuf, rh_dma_attr_buf);
		gcp->gc_dma_attr_txbuf.dma_attr_align  = gcp->gc_tx_buf_align+1;
		gcp->gc_dma_attr_txbuf.dma_attr_sgllen = gcp->gc_tx_max_frags;
		STRUCT_COPY(gcp->gc_dma_attr_rxbuf, rh_dma_attr_buf);
		gcp->gc_dma_attr_rxbuf.dma_attr_align  = gcp->gc_rx_buf_align+1;
		gcp->gc_dma_attr_rxbuf.dma_attr_sgllen = gcp->gc_rx_max_frags;


		/* timeout parameters */
		gcp->gc_tx_timeout = GEM_TX_TIMEOUT;
		gcp->gc_tx_timeout_interval = GEM_TX_TIMEOUT_INTERVAL;

		/* flow control */
		gcp->gc_flow_control = FLOW_CONTROL_NONE;
		if (IS_RHINE2_OR_LATER(revid)) {
			/* rx_pause implies symmetric */
			gcp->gc_flow_control = FLOW_CONTROL_RX_PAUSE;
		}

		/* MII timeout parameters */
		gcp->gc_mii_link_watch_interval = GEM_LINK_WATCH_INTERVAL;
		gcp->gc_mii_an_watch_interval = ONESEC/5;
		gcp->gc_mii_reset_timeout    = MII_RESET_TIMEOUT;
		gcp->gc_mii_an_timeout       = MII_AN_TIMEOUT;
		gcp->gc_mii_an_wait          = 0; /* was (25*ONESEC)/10 */
		gcp->gc_mii_linkdown_timeout = MII_LINKDOWN_TIMEOUT; 

		/* workaround for DAVICOM PHY */
		gcp->gc_mii_an_delay        = ONESEC/10;
		gcp->gc_mii_linkdown_action = MII_ACTION_RESET;
		gcp->gc_mii_dont_reset      = FALSE;

		/* I/O methods */

		/* mac operation */
		gcp->gc_attach_chip = &rh_attach_chip;
		gcp->gc_reset_chip = &rh_reset_chip;
		gcp->gc_init_chip  = &rh_init_chip;
		gcp->gc_start_chip = &rh_start_chip;
		gcp->gc_stop_chip  = &rh_stop_chip;
		gcp->gc_multicast_hash = &rh_mcast_hash;
		gcp->gc_set_rx_filter = &rh_set_rx_filter;
		gcp->gc_set_media = &rh_set_media;
		gcp->gc_get_stats = &rh_get_stats;
		gcp->gc_interrupt = &rh_interrupt;

		/* descriptor operation */
		gcp->gc_tx_desc_write = &rh_tx_desc_write;
		gcp->gc_rx_desc_write = &rh_rx_desc_write;

		gcp->gc_tx_desc_stat = &rh_tx_desc_stat;
		gcp->gc_rx_desc_stat = &rh_rx_desc_stat;
		gcp->gc_tx_desc_init = &rh_tx_desc_init;
		gcp->gc_rx_desc_init = &rh_rx_desc_init;
		gcp->gc_tx_desc_clean = &rh_tx_desc_clean;
		gcp->gc_rx_desc_clean = &rh_rx_desc_clean;
		gcp->gc_get_packet    = &gem_get_packet_default;

		/* mii operations */
		gcp->gc_mii_init = &gem_mii_init_default;
		gcp->gc_mii_config = &gem_mii_config_default;
		gcp->gc_mii_sync = &rh_mii_sync;
		gcp->gc_mii_read = &rh_mii_read;
		gcp->gc_mii_write = &rh_mii_write;
		gcp->gc_mii_tune_phy = NULL;

		lp = (struct rh_dev *)
			kmem_zalloc(sizeof(struct rh_dev), KM_SLEEP);
		lp->revid = revid;

		dp = gem_do_attach(dip, gcp, base_mm, &ha_mm, lp, sizeof(*lp));

		kmem_free(gcp, sizeof(*gcp));
		if (dp == NULL) {
			goto  err_free_mem;
		}

#ifdef OPT_TX_INTR
		if (IS_RHINE1(lp->revid)) {
			if (ddi_add_softintr(dp->dip,
					DDI_SOFTINT_LOW, &lp->soft_id,
					NULL, NULL,
					(uint_t (*)(caddr_t))&rh_update_imr,
					(caddr_t)dp) != DDI_SUCCESS) {
				goto  err_free_mem;
			}
		}
#endif
		return DDI_SUCCESS;

err_free_mem:
		kmem_free(lp, sizeof(struct rh_dev));
err:
		pci_config_teardown(&conf_handle);
		return DDI_FAILURE;
	}
	return DDI_FAILURE;
}

static int
rhdetach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
#ifdef OPT_TX_INTR
	struct gem_dev  *dp;
	gld_mac_info_t  *macinfo;
	struct rh_dev	*lp;

	macinfo = (gld_mac_info_t *)ddi_get_driver_private(dip);
	dp = (struct gem_dev *)macinfo->gldm_private;
	lp = (struct rh_dev *)dp->private;
#endif
	if (cmd == DDI_DETACH) {
#ifdef OPT_TX_INTR
		if (IS_RHINE1(lp->revid)) {
			ddi_remove_softintr(lp->soft_id);
		}
#endif
		return gem_do_detach(dip);
	}
	return DDI_FAILURE;
}

/* ======================================================== */
/*
 * OS depend (loadable streams driver) routine
 */
/* ======================================================== */
static	struct module_info rhminfo = {
	0,			/* mi_idnum */
	"rh",			/* mi_idname */
	0,			/* mi_minpsz */
	ETHERMTU,		/* mi_maxpsz */
	TX_BUF_SIZE * ETHERMAX,	/* mi_hiwat */
	1,			/* mi_lowat */
};

static	struct qinit rhrinit = {
	(int (*)()) NULL,	/* qi_putp */
	gem_rsrv,		/* qi_srvp */
	gem_open,		/* qi_qopen */
	gem_close,		/* qi_qclose */
	(int (*)()) NULL,	/* qi_qadmin */
	&rhminfo,		/* qi_minfo */
	NULL			/* qi_mstat */
};

static	struct qinit rhwinit = {
	gem_wput,		/* qi_putp */
	gem_wsrv,		/* qi_srvp */
	(int (*)()) NULL,	/* qi_qopen */
	(int (*)()) NULL,	/* qi_qclose */
	(int (*)()) NULL,	/* qi_qadmin */
	&rhminfo,		/* qi_minfo */
	NULL			/* qi_mstat */
};

static struct streamtab	rh_info = {
	&rhrinit,	/* st_rdinit */
	&rhwinit,	/* st_wrinit */
	NULL,		/* st_muxrinit */
	NULL		/* st_muxwrinit */
};

static	struct cb_ops cb_rh_ops = {
	nulldev,	/* cb_open */
	nulldev,	/* cb_close */
	nodev,		/* cb_strategy */
	nodev,		/* cb_print */
	nodev,		/* cb_dump */
	nodev,		/* cb_read */
	nodev,		/* cb_write */
	nodev,		/* cb_ioctl */
	nodev,		/* cb_devmap */
	nodev,		/* cb_mmap */
	nodev,		/* cb_segmap */
	nochpoll,	/* cb_chpoll */
	ddi_prop_op,	/* cb_prop_op */
	&rh_info,	/* cb_stream */
	D_NEW|D_MP	/* cb_flag */
};

static	struct dev_ops rh_ops = {
	DEVO_REV,	/* devo_rev */
	0,		/* devo_refcnt */
	gem_getinfo,	/* devo_getinfo */
	nulldev,	/* devo_identify */
	nulldev,	/* devo_probe */
	rhattach,	/* devo_attach */
	rhdetach,	/* devo_detach */
	nodev,		/* devo_reset */
	&cb_rh_ops,	/* devo_cb_ops */
	NULL,		/* devo_bus_ops */
	ddi_power	/* devo_power */
};

static struct modldrv modldrv = {
	&mod_driverops,	/* Type of module.  This one is a driver */
	ident,
	&rh_ops,	/* driver ops */
};

static struct modlinkage modlinkage = {
	MODREV_1, &modldrv, NULL
};

/* ======================================================== */
/*
 * _init : done
 */
/* ======================================================== */
int
_init(void)
{
	int 	status;

	DPRINTF(2, (CE_CONT, "!rh: _init: called"));
	status = mod_install(&modlinkage);

	return (status);
}

/*
 * _fini : done
 */
int
_fini(void)
{
	int	status;

	DPRINTF(2, (CE_CONT, "!rh: _fini: called"));
	status = mod_remove(&modlinkage);
	return (status);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}
